import torch


def log_likelihood(nbhd_means, feature_means, feature_vars, k, past_comps = []):
    # given neighborhood expression data, construct likelihood function
    # should work in pytorch
    n_samples = nbhd_means.shape[0]

    nbhd_means = torch.tensor(nbhd_means).double()
    feature_means = torch.tensor(feature_means).double()
    feature_vars = torch.tensor(feature_vars).double()

    def f(coefs):
        coefs_orth = coefs
        for i,comp in enumerate(past_comps):
            #project orthogonally
            coefs_orth = coefs_orth - torch.dot(coefs_orth, comp) * comp

        observed_means = torch.matmul(nbhd_means, torch.reshape(coefs_orth, (-1, 1)))

        # mean and variance of metagene defined by coef, under
        # independent gene hypothesis
        # variance gets scaled down due to averaging.
        theoretical_mean = torch.dot(feature_means, coefs_orth)
        theoretical_var = torch.div(torch.dot(torch.pow(coefs_orth, 2), feature_vars), float(k))

        #         print(theoretical_mean)
        #         print(theoretical_var)

        result = (-1 * n_samples / 2.) * torch.log(theoretical_var) - torch.div(
            torch.sum(torch.pow(observed_means - theoretical_mean, 2)),
            2 * theoretical_var)
        return (result)

    return (f)



from .utils import to_sparse_adjacency
from scipy.stats import rankdata
import numpy as np
from scipy.sparse import csr_matrix, hstack, issparse
from scipy.stats import norm, binom, ttest_ind_from_stats
import sys


class Scorer:
    def __init__(self, corrector=None, seed=None, verbose=False):
        self.corrector = corrector
        self.seed = seed
        self.verbose = verbose

    def score(self, X, nbhds, nn_matrix = None):
        k = len(nbhds[0])
        # if self.n_tests == 'auto':  # compute by bootstrapping
        #     self.n_tests = 1
        #     self.n_tests = bootstrapped_ntests(X, scorer=self, return_all=False, seed=self.seed)
        pass


class Tf_idfScorer(Scorer):

    def score(self, X, nbhds = None):
        #nbhds is ignored; just TF-IDF transform the data.
        if not issparse(X):
            X = csr_matrix(X)

        X_binarized = (X>0).astype("float")
        counts = X_binarized.sum(axis=0).A.flatten()
        idfs = np.ones(len(counts))
        idfs[counts>0] = np.log(X.shape[0]/counts[counts>0])

        scores = X.multiply(idfs)
        return(scores)

class BinomialScorer(Scorer):
    def __init__(self, corrector=None, seed=None, verbose=False, scaled=False,
                 fast_version=True, max_bins=float('inf')):
        super().__init__(corrector, seed, verbose)
        self.fast_version = fast_version
        self.max_bins = max_bins
        self.scaled = scaled
        self.verbose = verbose

    def get_binom_scores(self, gene_probs, k, max_bins=500, verbose=True, scaled=False,
                     error_rate=1.0):
        # compute binom_test(x, k, p) for all x in 1:k and all p in gene_probs
        # if scaled, put more density in lower probabilities
        if max_bins < float('inf'):
            # split interval into max_bins bins
            if self.scaled:
                if verbose:
                    print('using scaled bins')
                min_prob = np.min(gene_probs[gene_probs > 0])
                precision = (min_prob) ** (1. / float(max_bins))  # fold accuracy in probability approx
                splits = precision ** (max_bins - (np.arange(max_bins)))
            else:
                splits = np.arange(0, 1, 1 / max_bins)[1:]  # don't allow a gene to have 0 prob...

            # bins each gene gets put in, based on its global probability
            gene_bins = np.array([np.argmin(np.abs(gene_probs[j] - splits)) for j in range(len(gene_probs))])

            binom_scores = np.zeros((len(splits), k + 1))

            for i, p in enumerate(splits):
                if verbose:
                    print('\r computing binom scores for bin {}/{}'.format(i, len(splits)), end=' ')

                # #print('using fast scores')
                # if fast_scores:
                signs = [1 if p * k <= j else -1 for j in np.arange(k + 1)]
                pmfs = binom.pmf(np.arange(k + 1), k, p)
                orderer = np.argsort(pmfs)

                pvals = np.cumsum(pmfs[orderer])
                if self.corrector is not None:
                    pvals = self.corrector.correct(pvals)

                pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen
                binom_scores[i, orderer] = np.array(signs)[orderer] * -1 * np.log(pvals)

        else:  # compute for all gene probs
            binom_scores = np.zeros((len(gene_probs), k + 1))
            gene_bins = np.array(range(len(gene_probs)))

            for i in range(len(gene_probs)):
                p = gene_probs[i]

                if verbose:
                    print('\r computing binom scores for genes {}/{}'.format(i, len(gene_probs)), end=' ')
                if gene_probs[i] == 0 or gene_probs[i] == 1:  # no expression, no score
                    continue

                signs = [1 if p * k <= j else -1 for j in np.arange(k + 1)]
                pmfs = binom.pmf(np.arange(k + 1), k, p)
                orderer = np.argsort(pmfs)

                pvals = np.cumsum(pmfs[orderer])

                if self.corrector is not None:
                    pvals = self.corrector.correct(pvals)
                # pvals = 1-np.power(1-pvals, n_tests)

                pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen

                binom_scores[i, orderer] = np.array(signs)[orderer] * -1 * np.log(pvals)
        return (gene_bins, binom_scores)

    def score(self, X, nbhds, nn_matrix=None, binom_scores=None, gene_bins=None):
        super().score(X, nbhds)
        k = len(nbhds[0])
        gene_probs = np.array((X > 0).sum(axis=0) / float(X.shape[0])).flatten()
        nbhd_probs = np.zeros(X.shape)
        wts = np.zeros((len(nbhds), X.shape[1]))  # too big for large data

        X = csr_matrix((X > 0).astype('float'))  # convert to sparse binarized matrix
        if binom_scores is None or gene_bins is None:
            gene_bins, binom_scores = self.get_binom_scores(gene_probs, k)
        if self.fast_version:
            # compute significance of gene expression in each cell's neighborhood
            # first convert neighborhood to sparse matrix

            if nn_matrix is None:
                data = np.ones(np.sum([len(x) for x in nbhds]))
                col_ind = [item for sublist in nbhds for item in sublist]
                row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]
                # sparse adjacency matrix of NN graph
                nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

            # get gene expressions within each neighborhood; this matrix may be less sparse
            nbhd_exprs = (nn_matrix * X).astype('int').todense()

            # # extract locations and values of nonzero nbhd expressions.
            # rows, cols = nbhd_exprs.nonzero()
            # exprs = nbhd_exprs.data

            # apply binomial scores
            rows, cols = np.indices((len(nbhds), X.shape[1]))
            rows = rows.flatten()
            cols = cols.flatten()

            wts = binom_scores[gene_bins[cols], np.array(nbhd_exprs[rows, cols]).flatten()].reshape(
                (len(nbhds), X.shape[1]))

        else:
            for i in range(len(nbhds)):
                if self.verbose:
                    if i < len(nbhds) - 1:
                        print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         ')
                    else:
                        print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         \n')

                nnbhd = X[nbhds[i], :]

                nbhd_size = len(nbhds[i])
                nbhd_gene_counts = np.array((nnbhd > 0).sum(axis=0)).flatten()

                nbhd_probs[i, :] = nbhd_gene_counts / nbhd_size

                if self.max_bins < float('inf'):
                    # look up the binomial score in the nearest bins

                    # gene_scores = [binom_scores[gene_bins[j], count] for j, count in enumerate(nbhd_gene_counts)]
                    gene_scores = binom_scores[gene_bins, nbhd_gene_counts]

                else:
                    gene_scores = [binom_scores[j, count] for j, count in enumerate(nbhd_gene_counts)]

                wts[i, :] = gene_scores
                # expected_vals = nbhd_size * gene_probs
                # wts[i, :] = -1*np.log(gene_scores) * (2*(nbhd_gene_counts > expected_vals)-1)
        return(csr_matrix(wts))

class TScorer(Scorer):
    def __init__(self, corrector=None, seed=None, verbose=False):
        super().__init__(corrector, seed, verbose)
        self.verbose =  verbose
        self.seed = seed
        self.corrector = corrector

    def score(self, X, nbhds, nn_matrix = None):
        if not issparse(X):
            X = csr_matrix(X)

        k = len(nbhds[0])
        if nn_matrix is None:
            data = np.ones(np.sum([len(x) for x in nbhds]))
            col_ind = [item for sublist in nbhds for item in sublist]
            row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

            # sparse adjacency matrix of NN graph
            nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

        # get mean gene expressions within each neighborhood; this matrix may be less sparse
        mean_nbhd_exprs = (nn_matrix * X).astype('int').multiply(1 / nn_matrix.sum(axis=1)).tocsr()

        vars = np.zeros((len(nbhds), X.shape[1]))
        for i in range(len(nbhds)):  # gotta go cell by cell
            nbrs = np.array(nbhds[i]).flatten()
            gene_diffs = np.power((X[nbrs, :].todense() - mean_nbhd_exprs[i, :].todense()),
                                  2)  # diffs of gene expression
            vars[i, :] = gene_diffs.mean(axis=0)
        vars = csr_matrix(vars)

        global_means = np.tile(X.mean(axis=0), (len(nbhds), 1))

        # sign is pos if mean is higher, negative otherwise.
        signs = 2 * (mean_nbhd_exprs.todense() >= global_means).astype('int') - 1

        global_var = np.tile(np.var(X.todense(), axis=0), (len(nbhds), 1))
        nobs_global = np.tile(X.shape[0], (len(nbhds), X.shape[1]))
        nobs_local = np.tile(k, (len(nbhds), X.shape[1]))

        wts = ttest_ind_from_stats(mean1=mean_nbhd_exprs.todense().flatten(),
                                   std1=np.array(np.sqrt(vars.todense()).flatten()),
                                   nobs1=np.array(nobs_local).flatten(),
                                   mean2=np.array(global_means).flatten(),
                                   std2=np.array(np.sqrt(global_var)).flatten(),
                                   nobs2=np.array(nobs_global).flatten()).pvalue.reshape((len(nbhds), X.shape[1]))

        np.nan_to_num(wts, copy=False, nan=1.0)  # nans become pval 1

        wts[wts == 0] = sys.float_info.min  # remove zeros

        if self.corrector is not None:
            wts = self.corrector.correct(wts)

        wts = -1 * np.log(wts)  # convert to info

        np.nan_to_num(wts, copy=False, nan=1.0)  # nans become pval 1

        wts = np.multiply(signs, wts)  # negative if underexpressed

        return (csr_matrix(wts))

class WilcoxonScorer(Scorer):
    def __init__(self, corrector = None, seed=None, verbose=False):
        super().__init__(corrector, seed, verbose)

    def score(self, X, nbhds, nn_matrix=None):
        if not issparse(X):
            X = csr_matrix(X)

        k = len(nbhds[0])

        super().score(X, nbhds)  # handle multi-test factor determination
        # Wilcoxon rank sum testa
        # overall_exprs = X.todense().transpose().tolist()

        n_genes = X.shape[1]

        if nn_matrix is None:
            nn_matrix = to_sparse_adjacency(nbhds, n_cells=X.shape[0])

        wts = rankdata(X.todense(), axis=0)  # gene rankings
        wts = nn_matrix @ wts  # nbhd_ranksums; only want to store one big matrix

        n1 = k
        n2 = X.shape[0] - k
        sd = np.sqrt(n1 * n2 * (n1 + n2 + 1) / 12.0)
        meanrank = n1 * n2 / 2.0

        # #sign is pos if mean rank is higher than average, negative otherwise.
        # signs = 2*(wts >= meanrank).astype('int') - 1

        wts = wts - ((n1 * (n1 + 1)) / 2.0)  # calc U for x, u1

        is_neg = (wts < meanrank)  # remember where it was negative

        wts = np.maximum(wts, n1 * n2 - wts)  # bigu

        wts = ((wts - meanrank) / sd)  # z values

        wts = 2 * norm.sf(np.abs(wts))  # p values

        if self.corrector is not None:
            wts = self.corrector.correct(wts)

        wts = -1 * np.log(wts)  # convert to info scores

        # sign them
        wts[is_neg] *= -1

        return(csr_matrix(wts))

class ChunkedScorer(Scorer):

    def __init__(self, base_scorer, chunk_size=1000, max_size=None):

        self.verbose = base_scorer.verbose
        self.base_scorer = base_scorer
        self.chunk_size = chunk_size
        self.max_size = max_size

    def score(self, X, nbhds, **kwargs):

        #amortize this over all chunks.
        nn_matrix = to_sparse_adjacency(nbhds, n_cells=X.shape[0])

        if self.max_size is not None:
            # set chunksize so that at most maxsize floats are stored in a dense matrix at once
            chunksize = np.ceil(float(self.max_size) / X.shape[0]).astype('int')
        else:
            chunksize = self.chunk_size

        n_genes = X.shape[1]
        chunk_ends = [0] + list(np.arange(chunksize, n_genes, chunksize))
        chunk_ends.append(n_genes)
        gene_idxs = np.array(list(range(n_genes)))
        gene_chunks = [np.array(gene_idxs[chunk_ends[i]:chunk_ends[i + 1]]) for i in range(len(chunk_ends) - 1)]

        wt_blocks = [] # list of sparse matrices to concatenate horizontally

        for i,chunk in enumerate(gene_chunks):
            if self.verbose:
                print('chunk {}/{}'.format(i+1,len(gene_chunks)), end='\r')
            X_chunk = X[:,chunk]

            wt_blocks.append(self.base_scorer.score(X_chunk, nbhds=nbhds, nn_matrix=nn_matrix, **kwargs))

        return (csr_matrix(hstack(wt_blocks)))




"""a bunch of functions to generate fake data"""
import numpy as np
from scipy.sparse import csr_matrix
import scanpy as sc



def pancakes(n_pts=100, n_cor=2, n_anticor=2, n_noise=2, minor_fraction=0.5, random_fraction=0.5, marker_fraction=0.9,
             noise_level=0.1, dropout_clust=True):
    # split = [0] * int(np.ceil(n_pts * (1 - minor_fraction))) + [1] * int(np.floor(n_pts * minor_fraction))

    markers_on = np.random.choice([0, 1], size=(np.floor(n_pts * minor_fraction).astype('int'), n_cor),
                                  p=[1 - marker_fraction, marker_fraction])
    markers_off = np.random.choice([0, 1], size=(np.ceil(n_pts * (1 - minor_fraction)).astype('int'), n_cor),
                                   p=[1 - noise_level, noise_level])

    split_dims = np.vstack((markers_on, markers_off))
    # split_dims = np.tile(split, (n_anticor,1))

    normal_off = np.random.choice([0, 1], size=(np.floor(n_pts * minor_fraction).astype('int'), n_cor),
                                  p=[marker_fraction, 1 - marker_fraction])
    normal_on = np.random.choice([0, 1], size=(np.ceil(n_pts * (1 - minor_fraction)).astype('int'), n_cor),
                                 p=[noise_level, 1 - noise_level])

    split_dims = np.vstack((markers_on, markers_off))
    off_dims = np.vstack((normal_on, normal_off))

    # off =  [0]*int(np.floor(n_pts*minor_fraction)) + [1]*int(np.ceil(n_pts *(1-minor_fraction)))
    # off_dims = np.tile(off, (n_cor, 1)).transpose()

    noisy_dims = np.random.choice([0, 1], size=(n_noise, n_pts), p=[1 - random_fraction, random_fraction]).transpose()
    if dropout_clust:
        mat = np.concatenate((split_dims, off_dims, noisy_dims), axis=1)
    else:
        mat = np.concatenate((split_dims, noisy_dims), axis=1)
    # print(mat.shape)

    adata = sc.AnnData(csr_matrix(mat))

    labs = np.array(['normal'] * adata.shape[0])
    labs[:np.floor(n_pts * minor_fraction).astype('int')] = 'rare'
    if dropout_clust:
        labs[-1 * np.floor(n_pts * minor_fraction).astype('int'):] = 'dropout'
    # print(len(labs))

    adata.obs['label'] = labs
    # adata.obs['label'] = [str(x) for x in np.array(split)+np.array(off)]

    return adata



import scanpy as sc
from sklearn.neighbors import NearestNeighbors
from .score import info_score, bootstrapped_ntests
import numpy as np
from scipy.sparse import csr_matrix
from scipy import sparse
from .scorers import Scorer, WilcoxonScorer, BinomialScorer, ChunkedScorer, TScorer
from .correctors import FWERCorrector
from .embedders import SCAEmbedder
from .connectors import MetricConnector, Connector

def reduce(X, n_comps=50, iters=1, nbhds=None,
           nbhd_size=15,  metric='euclidean', model='wilcoxon',
           keep_scores=False, keep_loadings=False, keep_all_iters=False, verbose=False, n_tests = 'auto',
           seed=10,  chunk_size=None, **kwargs):
    """Compute an SCA reduction of the input data

    :param X: (num cells)x(num_genes)-sized array or sparse matrix to be dimensionality-reduced.
    :type X: numpy.ndarray | scipy.spmatrix
    :param n_comps: Desired dimensionality of the reduction. Default 50
    :type n_comps: int
    :param iters: Number of iterations of SCA. More iterations usually strengthens signal, stabilizing around 3-5
    :type iters: int
    :param nbhd_size: Size of neighborhoods used to assess the local expression of a gene. Should be smaller than the smallest subpopulation; default is 15.
    :type nbhd_size: int
    :param model: Model used to test for local enrichment of genes, used to compute information scores. One of ["wilcoxon","binomial","ttest"], default "wilcoxon" (recommended).
    :type model: str
    :param nbhds: Optional - if k-neighborhoods of points are already determined, they can be specified here as a (num_cells)*k array or list. Otherwise, they will be computed from the PCA embedding. Default None
    :type nbhds: numpy.ndarray | list
    :param metric: Metric used to compute k-nearest neighbor graphs for SCA score computation. Default "euclidean". See sklearn.neighbors.DistanceMetric for list of choices.
    :type metric: str
    :param keep_scores: if True, keep and return the information score matrix. Default False.
    :type keep_scores: bool
    :param keep_loadings: If True, returns loadings of each gene in each metagene as a dense matrix. Default False.
    :type keep_loadings: bool
    :param verbose: If True, print progress. Default False
    :type verbose: bool
    :param n_tests: Effective number of independent genes per cell, use for FWER multiple testing correction. Set to "auto" to automatically determine by bootstrapping. Default "auto".
    :type n_tests: str | int
    :param kwargs: Other arguments to be passed to the chosen scorer
    :return: If return_scores or return_loadings are both false, a (n cells)x(n_comps)-dimensional array of reduced features. Otherwise, a dictionary with keys 'reduction', 'scores' and/or 'loadings'.
    :rtype: numpy.ndarray | dict

    """

    if n_tests == 'auto':
        n_tests = bootstrapped_ntests(X, model=model, k=nbhd_size, seed=seed)
        if verbose:
            print('multi-testing correction for {} features'.format(n_tests))

    corrector = FWERCorrector(n_tests=n_tests)

    if issubclass(type(model), Scorer):
        # can directly specify model or give a string to construct one.
        scorer = model
    elif model == 'wilcoxon':
        scorer = WilcoxonScorer(corrector=corrector, verbose=verbose, **kwargs)
    elif model == 'binomial':
        scorer = BinomialScorer(corrector=corrector, verbose=verbose, **kwargs)
    elif model == 'ttest':
        scorer = TScorer(corrector=corrector, verbose=verbose, **kwargs)
    else:  # break
        assert False, 'scorer not found' #TODO fix

    if chunk_size is not None:
        scorer = ChunkedScorer(base_scorer=scorer, chunk_size=chunk_size)

    if issubclass(type(metric), Connector):
        connector = metric
    else:
        connector = MetricConnector(n_neighbors=nbhd_size, metric=metric, include_self=True)

    embedder = SCAEmbedder(scorer=scorer, connector=connector, n_comps=n_comps, iters=iters)
    dimred = embedder.embed(X, keep_scores=keep_scores, keep_loadings=keep_loadings, keep_all_iters=keep_all_iters,
                            nbhds=nbhds)

    if not keep_scores and not keep_loadings and not keep_all_iters:
        return dimred

    result = {}
    if not keep_all_iters:
        result['reduction'] = dimred  # store result
    else:
        result.update({('reduction_'+str(i)):j for i,j in embedder.embedding_dict.items()})
    if keep_scores:
        result['scores'] = embedder.scores
    if keep_loadings:
        result['loadings'] = embedder.loadings

    return result

def reduce_scanpy(adata, keep_scores=False, keep_loadings=True, keep_all_iters=False, layer=None, key_added='sca',
                  iters=1, model='wilcoxon',**kwargs):
    """
    Compute an SCA reduction of the given dataset, stored as a scanpy AnnData.

    :param adata: AnnData object containing single-cell transcriptomic data to be reduced
    :type adata: scanpy.AnnData
    :param keep_scores: if True, stores information score matrix in adata.layers[key_added+'_score']. Default False.
    :type keep_scores: bool

    """

    if layer is not None:
        X = adata.layers[layer]
    else:
        X = adata.X


    dimred_info = reduce(X, keep_scores=keep_scores, keep_loadings=keep_loadings,
                         keep_all_iters=keep_all_iters, iters=iters, model=model, **kwargs)

    if not keep_scores and not keep_loadings and not keep_all_iters:
        adata.obsm['X_'+key_added] = dimred_info
    else:
        if keep_all_iters:
            for i in range(iters):
                adata.obsm['X_'+key_added + '_' + str(i + 1)] = dimred_info['reduction_{}'.format(i + 1)]
        else:
            adata.obsm['X_'+key_added] = dimred_info['reduction']

        if 'scores' in dimred_info:
            adata.layers[key_added + '_score'] = dimred_info['scores']
        if 'loadings' in dimred_info:
            adata.varm[key_added + '_loadings'] = dimred_info['loadings']

##### DEPRECATED, left in temporarily to ensure new method outputs match old method outputs. ######

def reduce_old(X, n_comps=50, iters=1, max_bins=float('inf'), fast_version=True, scaled_bins=False, nbhds=None,
           rep=None, nbhd_size=15, n_pcs=50, metric='euclidean', model='wilcoxon',
           keep_scores=False, keep_loadings=False, keep_all_iters=False, verbose=False, n_tests = 'auto',
           seed=10,  **kwargs):
    """
    DEPRECATED: please use reduce.

    :param X: (num cells)x(num_genes)-sized array or sparse matrix to be dimensionality-reduced. Should be nonnegative,
    with 0 indicating no recorded transcripts (required for binarization and binomial inference).
    :param n_comps: Desired dimensionality of the reduction.
    :param iters: Number of iterations. More iterations yi
    :param max_bins: Resolution of gene frequency measurements. If max_bins=inf (default), each global gene frequency is
    measured exactly. Otherwise, they are binned into max_bins bins, reducing the number of binomial tests required.
    :param fast_version: If true (default), runtime will be faster but memory usage will be higher. If false, will be slower with
    a smaller memory footprint.
    :param scaled_bins: If true, frequency bins will be unevenly spaced over the unit interval via a power law,
    ensuring accuracy to a constant factor. If false (default), bins evenly divide the unit interval.
    :param nbhd_size: Size of neighborhoods used to assess the local expression of a gene. Should be smaller than the
    smallest subpopulation; default is 15. Does not drastically affect the
    :param nbhds: Optional - if k-neighborhoods of points are already determined, they can be specified here as
    a (num_cells)*k array. Otherwise, they will be computed from the PCS.
    :param rep: Optional - existing featurization for determining k-neighborhoods of points. If not provided,
    k-neighborhoods will be computed in PCA space.
    :param n_pcs: If either rep or nbhds is provided, this parameter is ignored. Otherwise, the number of PCs used for computing
    the starting representation, from which the starting k-neighborhoods are derived.
    :param metric: if either rep or nbhds is provided, this parameter is ignored. Otherwise, the metric used in computing
    the initial k-neighborhoods
    :param keep_scores: if True, returns information scores for each gene/cell combo in a sparse matrix.
    :param keep_loadings: If True, returns loadings of each gene in each metagene as a dense matrix.
    :param verbose: If True, print progress.
    :param kwargs: Other arguments to be passed to info_score
    :return: If return_scores or return_loadings are both false, a (n cells)x(n_comps)-dimensional array
    of reduced features. Otherwise, a dictionary with keys 'reduction', 'scores' and/or 'loadings'.
    """
    if n_tests == 'auto':
        n_tests = bootstrapped_ntests(X, model=model, k=nbhd_size, seed=seed, **kwargs)
        if verbose:
            print('multi-testing correction for {} features'.format(n_tests))

    result = {}  # final result, if return_all_iters=True

    if nbhds is None:
        if rep is None:
            a, bt, c, d = sc.tl.pca(X, n_comps=n_pcs, return_info=True)
            rep =  X @ bt.transpose()

            #rep = sc.tl.pca(X, n_comps=n_pcs)

        nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric)
        nn.fit(rep)

        nbhds = nn.kneighbors(return_distance=False)

        # add points to their own neighborhoods
        nbhds = np.concatenate((np.array(range(X.shape[0])).reshape(-1, 1), nbhds), axis=1)

    for i in range(iters):
        if verbose:
            print('\niteration {}'.format(i + 1))

        if model != 'binomial': #don't return bin info
            infos = info_score(X, nbhds,
                                                        max_bins=max_bins,
                                                        return_bin_info=True,
                                                        fast_version=fast_version,
                                                        verbose=verbose,
                                                        scaled=scaled_bins,
                                                        model=model,
                                                        n_tests=n_tests,
                                                        **kwargs)
        elif i == 0:
            # keep binomial scores for future runs
            infos, gene_bins, binom_scores = info_score(X, nbhds,
                                                        max_bins=max_bins,
                                                        return_bin_info=True,
                                                        fast_version=fast_version,
                                                        verbose=verbose,
                                                        scaled=scaled_bins,
                                                        model=model,
                                                        n_tests=n_tests,
                                                        **kwargs)
        else:
            infos = info_score(X, nbhds,
                               max_bins=max_bins,
                               binom_scores=binom_scores,
                               gene_bins=gene_bins,
                               fast_version=fast_version,
                               verbose=verbose,
                               scaled=scaled_bins,
                               model=model,
                               n_tests=n_tests,
                               **kwargs)


        a, bt, c, d = sc.tl.pca(infos, n_comps=n_comps, return_info=True)

        # pca(infos, k=n_comps, raw=True, pc_iters=pc_iters)

        current_dimred = X @ bt.transpose()  # metagene expression values in X

        # if scale_scs:
        #     norms = infos @ csr_matrix(bt.transpose())
        #     norms = sparse.linalg.norm(norms, axis=0, ord=2)
        #     print(norms)
        #     print(current_dimred.shape)
        #     print(norms.shape)
        #     current_dimred = current_dimred * norms
        if keep_all_iters:
            result['reduction_{}'.format(i + 1)] = current_dimred

        # compute neighborhoods for next iteration, using current reduction
        if i < (iters - 1):
            nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric)
            nn.fit(current_dimred)

            nbhds = nn.kneighbors(return_distance=False)
            nbhds = np.concatenate((np.array(range(X.shape[0])).reshape(-1, 1), nbhds), axis=1)

    if not keep_scores and not keep_loadings and not keep_all_iters:
        return current_dimred

    if not keep_all_iters:
        result['reduction'] = current_dimred  # store result
    if keep_scores:
        result['scores'] = csr_matrix(infos)
    if keep_loadings:
        result['loadings'] = bt.transpose()

    return result


"""functions for evaluating embeddings and neighbor networks"""
import scanpy as sc
import numpy as np
import pandas as pd


def fidelity(adata, name, groupby='label'):
    """look in .obsp[name_connectivities] for connectivity data and see how well it preserves the given labelling"""

    labels = adata.obs[groupby].values

    adj = (adata.obsp[name+'_'*(len(name)>0)+'connectivities']>0)
    edgelist = zip(*adj.nonzero())

    good_edges=0
    bad_edges=0
    for i,j in edgelist:
        if labels[i] == labels[j]:
            good_edges += 1
        else:
            bad_edges += 1


    return float(good_edges)/float(good_edges + bad_edges)


def group_fidelity(adata, name, groupby='label', pairwise=False):
    """matrix containing the fraction of edges that stay within each group
    if pairwise=True, returns the fraction of edges between each pair of groups
    i.e. position (i,j) is the proportion of edges that land in group j, among
    all edges originating in group i"""

    classes = adata.obs[groupby].values
    G = (adata.obsp[name+'_'*(len(name)>0)+'connectivities']>0)

    class_list = np.unique(classes)

    result = np.zeros((len(class_list), len(class_list)))

    for i, c in enumerate(class_list):
        # print(i)
        # print(c)
        inds = np.where(np.array(classes) == c)[0]

        G_sub = G[inds, :]  # only look at vertices in c
        class_freqs = [0] * len(class_list)
        for j in range(len(inds)):
            row = G_sub[j, :].todense().tolist()[0]
            row = np.array(row).astype(int)

            # print(G_sub[j,:].tolist()[0])
            nbr_inds = np.where(row > 0)[0]
            # print(nbr_inds)
            nbr_classes = np.array([classes[x] for x in nbr_inds])

            for k, c2 in enumerate(class_list):
                class_freqs[k] += np.sum(nbr_classes == c2)

        # print(class_freqs)
        result[i, :] = class_freqs

    result = result / result.sum(axis=1)[:, None]

    result = pd.DataFrame(result)
    result.columns = class_list
    result.index = class_list

    if pairwise:
        return (result)
    else:
        diags = np.diag(result.values)
        return pd.DataFrame({'cluster':class_list, 'fidelity':diags, 'method':name})



def plot_umap(adata, name, n_neighbors=10, rerun=False, **kwargs):
    """looks in .uns[name] for neighbors info, and uses that to store and plot a UMAP"""

    if name not in adata.uns or rerun:
        sc.pp.neighbors(adata, use_rep='X_'+name, key_added=name, n_neighbors=n_neighbors)

    if 'X_'+name+'_umap' not in adata.obsm or rerun:
        sc.tl.umap(adata, neighbors_key=name)
        adata.obsm['X_{}_umap'.format(name)] = adata.obsm['X_umap']

    sc.pl.embedding(adata, basis='{}_umap'.format(name), neighbors_key=name, **kwargs)







def good_bad_edges(adata, name, groupby='label'):

    labels = adata.obs[groupby].values

    adj = (adata.obsp[name + '_' * (len(name) > 0) + 'connectivities'] > 0)
    edgelist = zip(*adj.nonzero())

    good_edges = []
    bad_edges = []

    for i, j in edgelist:
        if labels[i] == labels[j]:
            good_edges.append((i,j))
        else:
            bad_edges.append((i,j))

    return (good_edges, bad_edges)


import unittest
from shannonca.embedders import *
from shannonca.generate import pancakes
from shannonca.scorers import *
from shannonca.correctors import *
from shannonca.connectors import *
from shannonca.utils import add_nbrs_from_dists, add_nbrs
from shannonca.dimred import reduce, reduce_old
import scanpy as sc
import numpy as np


class test_connectors(unittest.TestCase):

    def __init__(self,*args, **kwargs):
        super(test_connectors, self).__init__(*args, **kwargs)
        self.seed = 20
        np.random.seed(self.seed)
        self.testData = pancakes(n_pts=1000, n_cor=10, n_anticor=10, n_noise=1000)

    def test_MultiResConnector(self):

        base_embedder = SCAEmbedder(scorer=WilcoxonScorer(corrector=FWERCorrector(n_tests=100)),
                                    connector=MetricConnector(metric='euclidean', n_neighbors=15),
                                    n_comps=5)

        connector = MultiResConnector(base_embedder=base_embedder, base_metric='euclidean')

        nbhds = connector.connect(self.testData.X, resolution=50, n_neighbors=10)

        add_nbrs(self.testData, nbhds, dists=nbhds, key_added='multires')
        sc.tl.umap(self.testData, neighbors_key='multires')
        sc.pl.umap(self.testData, edges=True, neighbors_key='multires')

        return(nbhds)



import unittest
from shannonca.embedders import *
from shannonca.generate import pancakes
from shannonca.scorers import *
from shannonca.correctors import *
from shannonca.connectors import *
from shannonca.dimred import reduce, reduce_old
import scanpy as sc
import numpy as np

class test_embedders(unittest.TestCase):

    def __init__(self,*args, **kwargs):
        super(test_embedders, self).__init__(*args, **kwargs)
        self.seed = 20
        np.random.seed(self.seed)
        self.testData = pancakes(n_pts=100, n_cor=10, n_anticor=10, n_noise=100).X

    def test_SVDEmbedder(self):
        embedder = SVDEmbedder(n_comps=50)
        embedding = embedder.embed(self.testData)
        self.assertTrue(embedding.shape == (self.testData.shape[0], 50))

    def test_SCAEmbedder(self):
        corrector = FWERCorrector(n_tests=10)
        scorer = WilcoxonScorer(corrector=corrector)
        connector = MetricConnector(n_neighbors=15, metric='euclidean', include_self=True)

        embedder = SCAEmbedder(scorer=scorer, connector=connector, n_comps=50, iters=5)

        result = embedder.embed(self.testData, keep_scores=True, keep_all_iters=True, keep_loadings=True)

        result_old = reduce_old(self.testData, n_comps=50, iters=5, n_tests=10, nbhd_size=15,
                            n_pcs=50, metric='euclidean', model='wilcoxon', keep_scores=True)

        # make sure it's the same as the old one.
        diff = np.max(np.abs(result-result_old['reduction']))
        self.assertTrue(diff <= 1e-10)

        print('hi')

        #they are not the same...

        # #make neighborhoods
        # adata_new = sc.AnnData(result)
        # sc.pp.neighbors(adata_new, metric='euclidean')
        # sc.tl.umap(adata_new)
        # sc.pl.umap(adata_new)
        #
        # adata_old = sc.AnnData(result_old)
        # sc.pp.neighbors(adata_old, metric='euclidean')
        # sc.tl.umap(adata_old)
        # sc.pl.umap(adata_old)
        #

    def test_reducers(self):
        #make sure new modular reducer matches old reducer in output.

        res_new = reduce(self.testData, n_comps=50, iters=5, n_tests=10, nbhd_size=10,
                         metric='euclidean', model='binomial', keep_scores=True,
                         keep_all_iters=True)

        res_old = reduce_old(self.testData, n_comps=50, iters=5, n_tests=10, nbhd_size=10,
                            n_pcs=50, metric='euclidean', model='binomial', keep_scores=True,
                             keep_all_iters=True)




from scanpy.tools import pca
from .correctors import FWERCorrector

class Embedder:
    """
    Base class for Embedders. Does nothing, acting as an "identity" embedder.
    """
    def __init__(self):
        """
        Constructor, extended by child classes.
        """
        pass

    def embed(self,X):
        """
        Returns the input data, unchanged.
        """
        return(X)

class SVDEmbedder(Embedder):
    """
    Embedder that projects data to its top right singular vectors.
    """

    def __init__(self, n_comps=50, svd_mat = None, **kwargs):
        """
        Constructor

        :param n_comps: Number of top singular vectors to keep. Defines output dimensionality.
        :type n_comps: int
        :param svd_mat: if provided, project the input to the right singular vectors of this matrix. If None (default), project to singular vectors of the input.
        :type svd_mat: numpy.ndarray | matrix | spmatrix
        :param kwargs: Additional arguments passed to scanpy.tl.pca
        """

        self.n_comps = n_comps
        self.kwargs = kwargs
        self.svd_mat = svd_mat

    def embed(self, X, keep_loadings=False):
        """
        Project X onto a set of top right singular vectors, and return the result

        :param X: Data to be reduced
        :type X: np.ndarray | matrix | spmatrix
        :param keep_loadings: Whether to keep the loading vectors. If True, the loading vectors are stored as self.loadings. Default False.
        :type keep_loadings: bool

        :return: A low-dimensional representation of X.
        :rtype: np.ndarray | matrix | spmatrix
        """
        if self.svd_mat is None:
            self.svd_mat = X
        a, bt, c, d = pca(self.svd_mat, n_comps=self.n_comps, return_info=True, **self.kwargs)

        if keep_loadings:
            self.loadings = bt.transpose()
        return X @ bt.transpose()



class SCAEmbedder(Embedder):
    def __init__(self, scorer, connector=None, n_comps=50, iters=1):
        self.n_comps = n_comps
        self.iters = iters
        self.scorer = scorer
        self.connector = connector
        self.embedding_dict = {}

    def embed(self, X, keep_scores=False, keep_loadings=False, keep_all_iters=False, nbhds=None, verbose=True):
        assert (self.connector is not None) or (nbhds is not None), "must specify connector or neighborhoods"

        if self.iters == 0:
            # base case: just do PCA

            embedder = SVDEmbedder(n_comps=self.n_comps)

            if keep_all_iters:
                self.embedding_dict = {}

            return embedder.embed(X)


        else:
            #make nbhds if not provided
            if nbhds is None or self.iters > 1:
                base_embedder = SCAEmbedder(scorer=self.scorer, connector=self.connector,
                                            n_comps=self.n_comps, iters=self.iters - 1)

                base_embedding = base_embedder.embed(X, keep_scores=False, keep_loadings=False,
                                                     keep_all_iters=keep_all_iters, nbhds=nbhds)
                self.embedding_dict = base_embedder.embedding_dict
                if verbose:
                    print('iteration {}'.format(self.iters))
                nbhds = self.connector.connect(base_embedding)
            else:
                self.embedding_dict = {}

            scores = self.scorer.score(X, nbhds) #compute info scores

            final_embedder = SVDEmbedder(n_comps=self.n_comps, svd_mat=scores)

            result = final_embedder.embed(X, keep_loadings=True)

            if keep_scores:
                self.scores = scores
            if keep_loadings:
                self.loadings = final_embedder.loadings
            if keep_all_iters:
                self.embedding_dict[self.iters] = result

            return(result)









import numpy as np

def taylor_exp(x, n):
    # taylor series of 1-(1-x)^n
    # use for multiple test correction if x is very small.
    return (n * x
            - 0.5 * np.power(x, 2) * (n - 1) * n
            + (1. / 6.) * np.power(x, 3) * (n - 2) * (n - 1) * n
            - (1. / 24) * np.power(x, 4) * (n - 3) * (n - 2) * (n - 1) * n)



class FWERCorrector:
    def __init__(self, n_tests=1):
        self.n_tests = n_tests

    def correct(self, pvals):
        if self.n_tests > 1:
            # use FWER to correct for testing many genes
            pvals[pvals > 1e-10] = 1 - np.power(1 - pvals[pvals > 1e-10], self.n_tests)
            pvals[pvals <= 1e-10] = taylor_exp(pvals[pvals <= 1e-10], self.n_tests)
        return(pvals)


from sklearn.metrics import f1_score
import numpy as np
from itertools import chain, combinations


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


# def cluster_f1_dict(labels, true):
#     labs = np.unique(labels)
#     true_labs = np.unique(true)
#
#     mapping_dict = {}
#     for lab in labs:
#         overlap_scores = []
#         for tlab in true_labs:
#             overlap = np.sum(np.logical_and(labels == lab, true_labs == tlab))/np.sum(labels == lab)
#             overlap_scores.append(overlap)
#         mapping_dict[lab] =


def cluster_f1_best(labels, true, return_labs=False):
    # 1. identify all clusters intersecting the target population
    labs = np.unique(labels)
    overlapping_labs = []
    subsets = []
    for lab in labs:
        overlap = (np.logical_and(labels == lab, true))
        if np.sum(overlap) == np.sum(labels == lab):  # strict subset; add automatically
            subsets.append(lab)
        elif np.sum(np.logical_and(labels == lab, true)) > 0:
            overlapping_labs.append(lab)

    # print('{} overlapping clusters'.format(len(overlapping_labs)))

    sets = list(powerset(overlapping_labs))
    print('hi')
    print(len(list(sets)))
    print('{} combinations to try'.format(len(list(sets))))

    max_set = subsets
    max_f1 = None
    for candidate in list(sets):
        s = tuple(subsets) + candidate
        # print(s)
        if len(s) == 0 and len(subsets) == 0:  # skip empty set
            continue
        if len(s) > 2:
            union = np.logical_or.reduce(tuple([labels == x for x in s]))
        else:
            union = np.array(labels == s[0])
        f1 = f1_score(true, union)

        if max_f1 is None or f1 > max_f1:
            max_f1 = f1
            max_set = s
    if return_labs:
        return (max_f1, max_set)
    else:
        return (max_f1)


def cluster_f1_better(labels, true, return_labs=False):
    labs = np.unique(labels)
    overlapping_labs = []
    subsets = []
    max_indicator = np.zeros(len(labels))
    max_f1 = 0
    max_clusts = []

    lab_scores = []
    #order sets by (true overlap) vs. (false overlap) proportion
    for lab in labs:
        tp = np.sum(np.logical_and(labels == lab, true))
        fp = np.sum(np.logical_and(labels == lab, np.logical_not(true)))

        score = tp/(tp+fp)
        lab_scores.append((score, lab))

    ordered_labs = [x[1] for x in sorted(lab_scores, reverse=True)] # labs ordered by true pos proportion

    for lab in ordered_labs:
        indicator = max_indicator + (labels == lab)
        f1 = f1_score(true, indicator)
        if f1 >= max_f1:
            max_indicator = indicator
            max_clusts.append(lab)
            max_f1 = f1_score(true, max_indicator)
        else:
            break

    if return_labs:
        return (max_f1, max_clusts)
    else:
        return max_f1



def cluster_f1(labels, true, return_labs=False):
    # given cluster labels, compute the best F1 score obtained by a combination of these clusters

    max_f1 = 0
    max_clusts = []
    labels = np.array(labels)

    # find cluster with highest F1 on its own
    labs = np.unique(labels)
    max_indicator = np.zeros(len(labels))
    for lab in labs:  # add all clusters that are subsets of true pop
        if np.sum(np.logical_and(labels == lab, true)) == np.sum(labels == lab):
            max_clusts.append(lab)
            max_indicator += (labels == lab)
            max_f1 = f1_score(true, max_indicator)

    # for lab in labs:
    #     indicator = (labels == lab)
    #     f1 = f1_score(true, indicator)
    #     if f1 > max_f1:
    #         max_clusts = [lab]
    #         max_indicator = indicator
    #         max_f1 = f1

    static = False  # did max_f1 change after cycling through the labs?
    while not static:
        static = True
        for lab in labs:
            if lab in max_clusts:
                continue

            indicator = max_indicator + (labels == lab)
            f1 = f1_score(true, indicator)
            if f1 > max_f1:
                static = False
                max_f1 = f1
                max_indicator = indicator
                max_clusts += [lab]

    if return_labs:
        return (max_f1, max_clusts)
    else:
        return max_f1


def ranking_f1(rankings, true, return_idxs=False):
    order = np.array(list(reversed(np.argsort(rankings))))
    f1s = np.zeros(len(rankings))

    for i in range(len(rankings)):
        # print('{}/{}'.format(i, len(rankings)), end='\r')
        indicator = np.zeros(len(rankings))
        indicator[order[:i]] = 1
        f1s[i] = f1_score(true, indicator)
    if return_idxs:
        return (np.max(f1s), indicator)

    return np.max(f1s)



from sklearn.neighbors import NearestNeighbors
import numpy as np
from scipy.sparse import csr_matrix
from sklearn.metrics import pairwise_distances

def knn_cover(X, k=50, metric='euclidean',n_covers=1, seed=10, **kwargs):
    sets = []
    np.random.seed(seed)
    for i in range(n_covers):
        is_uncovered = np.array([True]*X.shape[0])

        nn = NearestNeighbors(metric=metric, n_neighbors=k, **kwargs)
        nn.fit(X)
        while np.sum(is_uncovered) > 0:
            seed = np.random.choice(np.arange(X.shape[0])[is_uncovered])
            new_set = nn.kneighbors(X[seed,:].reshape(1,-1), return_distance=False)
            new_set =np.append(new_set, seed) # add self to neighborhood
            sets.append(new_set.flatten())
            is_uncovered[new_set.flatten()] = False
    return sets

def multires_dists(X, covering_sets, embedder, metric='euclidean', aggregator='max',  max_neighbors=float('inf'), **kwargs):
    # d(x,y) is maximum distance in any covering set PCA

    distance_matrices = []
    final_matrix = csr_matrix((X.shape[0], X.shape[0]))

    # compute sparse pairwise distance matrices of each covering set
    for i, s in enumerate(covering_sets):
        print('set {}/{}'.format(i, len(covering_sets)), end='\r')
        X_sub = X[s.flatten(), :]

        X_sub_dimred = embedder.embed(X_sub, verbose=False)

        if max_neighbors == float('inf'):

            dists_sub = pairwise_distances(X_sub_dimred, metric=metric)

            cols = np.array(s.tolist() * len(s))
            rows = np.array([[i] * len(s) for i in s]).flatten()

            new_mat = csr_matrix((dists_sub.flatten(), (rows, cols)), shape=(X.shape[0], X.shape[0]))

        else:
            nn = NearestNeighbors(n_neighbors=max_neighbors, metric=metric)
            nn.fit(X_sub_dimred)

            dists_sub = nn.kneighbors_graph()

            rows, cols = dists_sub.nonzero()

            # in overall dataset
            global_rows = np.array(s)[rows]
            global_cols = np.array(s)[cols]

            new_mat = csr_matrix((dists_sub.data, (global_rows, global_cols)), shape=(X.shape[0], X.shape[0]))

        distance_matrices.append(new_mat)

        if aggregator == 'max':
            final_matrix = final_matrix.maximum(new_mat)

    return final_matrix




from .dimred import reduce, reduce_scanpy
import numpy as np
import pandas as pd

def correct(Xs, feat_names=None, n_features=1000, **kwargs):
    if feat_names is None:
        feat_names = range(Xs[0].shape[1])

    sigs = np.zeros((len(Xs), Xs[0].shape[1]))
    for i, X in enumerate(Xs):
        scores = reduce(X, keep_scores=True, **kwargs)['scores']
        gene_sigs = np.array(abs(scores).mean(axis=0)).flatten()
        # print(gene_sigs)
        sigs[i, :] = gene_sigs

    gene_scores = np.mean(np.abs(sigs), axis=0)

    top_feats = np.argsort(gene_scores)[(-1 * n_features):]

    top_names = np.array(feat_names)[top_feats]

    corrected_Xs = [X[:, top_feats] for X in Xs]

    return (corrected_Xs, top_names)


def correct_scanpy(adata, batch_key='batch', n_features=1000, **kwargs):
    gene_sigs_dict = {}
    for i, batch in enumerate(np.unique(adata.obs[batch_key].tolist())):
        # print(batch)
        abatch = adata[adata.obs[batch_key].values.astype('str') == str(batch), :].copy()
        reduce_scanpy(abatch, keep_scores=True, keep_loadings=False, **kwargs)
        gene_sigs = np.array(np.mean(np.abs(abatch.layers['scalpel_score'].todense()), axis=0)).flatten()
        gene_sigs_dict[batch] = gene_sigs

    gene_sig_df = pd.DataFrame(gene_sigs_dict)
    gene_scores = np.mean(gene_sig_df.values, axis=1)
    most_sig_genes = adata.var.index[np.argsort(gene_scores)[(-1 * n_features):]]

    adata_corrected = adata[:, most_sig_genes].copy()
    return (adata_corrected, most_sig_genes)

# def correct_scanpy(adata, batch_key):



import unittest
from shannonca.generate import pancakes
from shannonca.scorers import  *
from shannonca.correctors import FWERCorrector
from sklearn.neighbors import NearestNeighbors
from shannonca.dimred import reduce
import seaborn as sns
import matplotlib.pyplot as plt
from shannonca.score import info_score

class test_scorers(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(test_scorers, self).__init__(*args, **kwargs)
        self.testData = pancakes(n_pts=100, n_cor=10, n_anticor=10, n_noise=100).X
        self.seed = 20
        nn = NearestNeighbors(metric='cosine', n_neighbors=10); nn.fit(self.testData)
        self.nbhds = nn.kneighbors(return_distance=False)

    def test_WilcoxonScorer(self):
        scorer = WilcoxonScorer(seed=self.seed, verbose=True, corrector=FWERCorrector(n_tests=10))
        scores = scorer.score(self.testData, self.nbhds)

        old_scores = info_score(self.testData, nbhds=self.nbhds, n_tests=10, model='wilcoxon',
                                chunk_size=10000)

        self.assertTrue(np.max(scores - old_scores) <= 1e-10)

        sns.heatmap(scores.todense(), cmap='coolwarm')
        plt.show()
        plt.savefig('figs/wilcoxon_pancake_scores.png')

    def test_BinomialScorer(self):
        scorer = BinomialScorer(seed=self.seed, verbose=True, corrector=FWERCorrector(n_tests=10))
        scores = scorer.score(self.testData, self.nbhds)
        scores_old = info_score(self.testData, self.nbhds, chunk_size=1000000, n_tests=10, model='binomial')
        self.assertTrue(np.max(scores - scores_old) <= 1e-10)

    def test_Tscorer(self):
        scorer = TScorer(seed=self.seed, verbose=True, corrector=FWERCorrector(n_tests=10))
        scores = scorer.score(self.testData, self.nbhds)
        scores_old = info_score(self.testData, self.nbhds, chunk_size=1000000, n_tests=10, model='ttest')
        self.assertTrue(np.max(scores-scores_old) <= 1e-10)

    def test_ChunkedScorer(self):
        base_scorer = WilcoxonScorer(seed=self.seed, verbose=True, corrector=FWERCorrector(n_tests=10))
        chunked_scorer_1 = ChunkedScorer(base_scorer=base_scorer, chunk_size=10)
        chunked_scorer_2 = ChunkedScorer(base_scorer=base_scorer, max_size=500)


        unchunked_scores = base_scorer.score(self.testData, self.nbhds).todense()
        fixedchunk_scores = chunked_scorer_1.score(self.testData, self.nbhds).todense()
        maxchunk_scores = chunked_scorer_2.score(self.testData, self.nbhds).todense()

        self.assertTrue((unchunked_scores==fixedchunk_scores).all())
        self.assertTrue((fixedchunk_scores==unchunked_scores).all())

    def test_Tf_idfScorer(self):
        scorer = Tf_idfScorer()
        scores = scorer.score(self.testData).todense()
        sns.heatmap(scores)









from .dimred import reduce, reduce_scanpy
from .generate import pancakes
from .integrate import correct, correct_scanpy
from .plotting import ordered_matrixplot
from .score import get_binom_scores, info_score
from .utils import metagene_loadings



from distutils.core import setup
from setuptools import find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
  name = 'shannonca', 
  packages = find_packages(exclude=['data']),  
  version = '0.0.7',   
  license='MIT',       
  description = 'Informative Dimensionality Reduction via Shannon Component Analysis',   
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Benjamin DeMeo',                   
  author_email = 'bdemeo@g.harvard.edu',      
  url = 'https://github.com/bdemeo/shannonca',  
  download_url =  'https://github.com/bendemeo/shannonca/archive/refs/tags/0.0.1.tar.gz',
  keywords = ['Shannon', 'Information', 'Dimensionality','reduction', 'single-cell','RNA'], 
  install_requires=[           
          'sklearn',
          'scipy',
          'numpy',
          'pandas',
          'scanpy'
      ]
    )



import numpy as np
#from .score import info_score
from scipy.sparse import issparse, csr_matrix
import scanpy as sc
from sklearn.neighbors import NearestNeighbors
import matplotlib.pyplot as plt
from scanpy.neighbors import _compute_connectivities_umap

def to_sparse_adjacency(nbhds, n_cells=None):
    if type(nbhds) is np.ndarray:
        nbhds = nbhds.tolist()

    # make nbhd adjacency matrix
    data = np.ones(np.sum([len(x) for x in nbhds]))

    col_ind = [item for sublist in nbhds for item in sublist]
    row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

    if n_cells is None:
        n_cells = np.max(col_ind)+1 # infer based on neighbor indices

    # sparse adjacency matrix of NN graph
    nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), n_cells))
    return(nn_matrix)

def dist_mat_to_nbhds(pairwise_dists, include_self=True, k=15):
    # given adjacency matrix, make a list of neighborhoods and their distances as recorded in the matrix.
    candidate_idxs = np.split(pairwise_dists.indices,
                              pairwise_dists.indptr[1:-1])  # list for each obs of other obs with recorded dists
    candidate_dists = np.split(pairwise_dists.data, pairwise_dists.indptr[1:-1])  # those distances

    new_idxs = [y[np.argsort(x)[:k]] for x, y in zip(candidate_dists, candidate_idxs)]
    new_dists = [np.sort(x)[:k] for x in candidate_dists]

    if include_self:
        new_idxs = [np.append(np.array([i]),idxs) for i, idxs in enumerate(new_idxs)]
        new_dists = [np.append(0,dist) for i, dist in enumerate(new_dists)]

    return((new_idxs, new_dists))





def metagene_loadings(data, n_genes=10, rankby_abs=False, key='sca'):
    """
    return the top-loaded genes for each Scalpel metagene
    :param data: AnnData or dict output from reduce_scanpy or reduce. Must contain loading data.
    :param n_genes: Number of top-loaded genes to record
    :param rankby_abs: If True, record the top genes by absolute value.
    :param key: Namespace of Scalpel data, if data is a scanpy object. Will look under data.varm[key+'_loadings'].
    :return: Dictionary mapping [sca component number]:{[top gene 1]:loading, [top_gene_2]:loading... [top_gene_k]:loading}
    """

    if type(data) is dict:
        loadings = data['loadings']
        var_names = np.array(list(range(loadings.shape[0])))
    else:
        loadings = data.varm[key + '_loadings']
        var_names = np.array(data.var_names)

    loadings /= np.sum(np.abs(loadings), axis=0)  # normalize to percent contribution
    loadings *= loadings.shape[0]  # fold enrichment over expected

    if rankby_abs:
        idxs = np.transpose(np.argsort(np.abs(loadings), axis=0)[(-1 * n_genes):, :]).tolist()
    else:
        idxs = np.transpose(np.argsort(loadings, axis=0)[(-1 * n_genes):, :]).tolist()

    genes = [var_names[np.array(x)] for x in idxs]
    scores = [loadings[x, i] for i, x in enumerate(idxs)]

    result = list(zip(genes, scores))
    result = {i: {'genes': k[np.argsort(-1 * v)], 'scores': v[np.argsort(-1 * v)]} for i, (k, v) in enumerate(result)}
    return result

def scramble_genes(X):
    if issparse(X):
        Xin = X.todense()
    else:
        Xin=X

    # from https://stackoverflow.com/questions/27486677/best-way-to-permute-contents-of-each-column-in-numpy
    ix_i = np.random.sample(Xin.shape).argsort(axis=0)
    ix_j = np.tile(np.arange(Xin.shape[1]), (Xin.shape[0], 1))

    Xin =  Xin[ix_i, ix_j]

    if issparse(X):
        return csr_matrix(Xin)
    else:
        return(Xin)

def binomial_background(X, nbhd_size=15, n_pcs=20, metric='cosine'):
    rep = sc.tl.pca(X, n_comps=n_pcs)

    nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric)
    nn.fit(rep)

    nbhds = nn.kneighbors(return_distance=False)

def make_umaps(adata, obsm_keys, n_neighbors=15, metric='cosine', **kwargs):
    for k in obsm_keys:
        print(k)
        sc.pp.neighbors(adata, use_rep='X_'+k, key_added=k, metric=metric, n_neighbors=n_neighbors)
        sc.tl.umap(adata, neighbors_key=k, **kwargs)
        adata.obsm['X_umap_{}'.format(k)] = adata.obsm['X_umap']

def plot_umaps(adata, umap_keys, plot_size=3, neighbors_keys=None, **kwargs):
    fig, axs = plt.subplots(1, len(umap_keys))
    if len(umap_keys)==1:
        axs = [axs]
    fig.set_size_inches(plot_size*len(umap_keys), plot_size)

    for i,k in enumerate(umap_keys):
        if neighbors_keys is None:
            key = None
        else:
            key = neighbors_keys[i]
        sc.pl.embedding(adata, basis='X_umap_{}'.format(k), s=20, ax = axs[i], show=False, neighbors_key=key, legend_loc=None, **kwargs)
        axs[i].set_title(k)
        axs[i].set_xlabel("UMAP 1")
        axs[i].set_ylabel("UMAP 2")
    fig.tight_layout()
    return fig

# from https://gist.github.com/sumartoyo/edba2eee645457a98fdf046e1b4297e4
def sparse_vars(a, axis=None):
    """ Variance of sparse matrix a
    var = mean(a**2) - mean(a)**2
    """
    a_squared = a.copy()
    a_squared.data **= 2
    return a_squared.mean(axis) - np.square(a.mean(axis))

def add_nbrs(adata, idxs, dists, key_added='new', add_self=True):
    # assumes same number of neighbors
    k = idxs.shape[1]

    if add_self:
        dists = np.concatenate((np.zeros((dists.shape[0], 1)), dists), axis=1)[:, :-1]
        idxs = np.concatenate((np.arange(idxs.shape[0]).reshape(-1, 1), idxs), axis=1)[:, :-1]

    dists, conns = _compute_connectivities_umap(idxs, dists, n_obs=adata.shape[0], n_neighbors=k)

    adata.obsp[key_added + '_distances'] = dists
    adata.obsp[key_added + '_connectivities'] = conns
    adata.uns[key_added] = {'connectivities_key': key_added + '_connectivities',
                            'distances_key': key_added + '_distances',
                            'params': {'method': 'umap'}}

def add_nbrs_from_dists(adata, pairwise_dists, k=10, **kwargs):
    # given a (sparse) pairwise distance matrix, add k-nearest neighbors connectivity info to adata.

    candidate_idxs = np.split(pairwise_dists.indices,
                              pairwise_dists.indptr[1:-1])  # list for each obs of other obs with recorded dists
    candidate_dists = np.split(pairwise_dists.data, pairwise_dists.indptr[1:-1])  # those distances

    new_idxs = [y[np.argsort(x)[:k]] for x, y in zip(candidate_dists, candidate_idxs)]
    new_dists = [np.sort(x)[:k] for x in candidate_dists]

    add_nbrs(adata, idxs=np.vstack(new_idxs), dists=np.vstack(new_dists), **kwargs)




from sklearn.neighbors import NearestNeighbors
import numpy as np
from .embedders import Embedder
from .lknn import multires_dists, knn_cover
from .utils import dist_mat_to_nbhds


class Connector:
    def __init__(self):
        pass

    def connect(self,X, **kwargs):
        # insert code to make nearest neighbor graphs
        # should return a LIST of neighborhoods
        pass


class MetricConnector(Connector):
    def __init__(self, n_neighbors=15, metric='euclidean', include_self=True):
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.include_self=include_self

    def connect(self, X, **kwargs):
        nn = NearestNeighbors(n_neighbors=self.n_neighbors, metric=self.metric, **kwargs)
        nn.fit(X)
        nbrs = nn.kneighbors(return_distance=False)


        if self.include_self:
            nbrs = np.concatenate((np.array(range(X.shape[0])).reshape(-1, 1), nbrs), axis=1)

        return(nbrs.tolist())

class MultiResConnector(Connector):
    def __init__(self, base_embedder, base_metric='euclidean', sub_embedder = None, sub_metric=None, verbose=False):
        self.base_embedder = base_embedder
        self.base_metric = base_metric
        self.verbose = verbose
        if sub_embedder is None:
            sub_embedder = base_embedder
        self.sub_embedder = sub_embedder
        if sub_metric is None:
            sub_metric = base_metric
        self.sub_metric = sub_metric
        self.sub_connector = sub_metric

    def connect(self, X, resolution, n_neighbors=15, aggregator='max', max_neighbors=float('inf'), n_covers=1,
                cover=None, return_dists=False, return_mat=False, **kwargs):
        # compute cover based on global embedding
        if cover is None:
            global_embedding = self.base_embedder.embed(X)
            cover = knn_cover(global_embedding, metric=self.base_metric, k=resolution, n_covers=n_covers)
        dist_mat = multires_dists(X, covering_sets=cover, embedder=self.sub_embedder,
                                  metric=self.sub_metric, aggregator=aggregator, max_neighbors=max_neighbors)

        if return_mat:
            return(dist_mat)
        nbhds, dists = dist_mat_to_nbhds(dist_mat, k=n_neighbors)
        if return_dists:
            return((nbhds, dists))
        else:
            return(nbhds)







from scipy.stats import binom
from scipy.sparse import csr_matrix, hstack
from scipy.stats import ttest_ind_from_stats, mannwhitneyu
from scipy.stats.distributions import norm
from scipy.stats.mstats import rankdata
from .utils import sparse_vars
from .scorers import *
import numpy as np

import sys
def taylor_exp(x, n):
    # taylor series of 1-(1-x)^n
    # use for multiple test correction if x is very small.
    return (n * x
            - 0.5 * np.power(x, 2) * (n - 1) * n
            + (1. / 6.) * np.power(x, 3) * (n - 2) * (n - 1) * n
            - (1. / 24) * np.power(x, 4) * (n - 3) * (n - 2) * (n - 1) * n)

def log1m_taylor(x):
    return(-1*x -np.power(x,2)/2 - np.power(x,3)/3 - np.power(x,4)/4 -np.power(x,5)/5 - np.power(x,6)/6)

def bootstrapped_ntests(X, trials=1000, k=15, model='binomial', return_all=False, seed=None, **kwargs):
    np.random.seed(seed)
    random_nbhds = np.random.choice(X.shape[0], size=(trials, k), replace=True)

    if issubclass(type(model), Scorer):
        # can directly specify model or give a string to construct one.
        scorer = model
    elif model == 'wilcoxon':
        scorer = WilcoxonScorer()
    elif model == 'binomial':
        scorer = BinomialScorer()
    elif model == 'ttest':
        scorer = TScorer()
    else:  # break
        assert False, 'scorer not found' #TODO fix

    nbhd_scores = scorer.score(X, nbhds=random_nbhds)
    #nbhd_scores = info_score(X, random_nbhds, n_tests=1, model=model, **kwargs)
    nbhd_ps = np.exp(-1 * np.abs(nbhd_scores.todense()))

    nbhd_minps = np.min(nbhd_ps, axis=1).A.flatten()

    sorted_minps = sorted(nbhd_minps)

    # empirical frequencies of low minimum p-values
    xvals = np.array(sorted_minps)
    yvals = np.arange(len(sorted_minps)) / len(sorted_minps)


    log1m_xvals = np.log(1 - xvals)
    log1m_yvals = np.log(1 - yvals)
    log1m_xvals[xvals<1e-10] = log1m_taylor(xvals[xvals<1e-10])

    xvals = log1m_xvals
    yvals = log1m_yvals

    # linear fit
    coef = np.mean(np.multiply(np.array(xvals), np.array(yvals))) / np.mean(np.power(np.array(xvals), 2))
    if return_all:
        return((xvals, yvals, np.ceil(coef).astype('int')))
    else:
        return(np.ceil(coef).astype('int'))

def get_binom_scores(gene_probs, k, max_bins=500, verbose=True, scaled=False, n_tests=50, multi_correct=True,
                     error_rate=1.0):
    # compute binom_test(x, k, p) for all x in 1:k and all p in gene_probs
    # if scaled, put more density in lower probabilities
    if max_bins < float('inf'):
        # split interval into max_bins bins

        if scaled:
            if verbose:
                print('using scaled bins')
            min_prob = np.min(gene_probs[gene_probs > 0])
            precision = (min_prob) ** (1. / float(max_bins))  # fold accuracy in probability approx
            splits = precision ** (max_bins - (np.arange(max_bins)))
        else:
            splits = np.arange(0, 1, 1 / max_bins)[1:]  # don't allow a gene to have 0 prob...

        # bins each gene gets put in, based on its global probability
        gene_bins = np.array([np.argmin(np.abs(gene_probs[j] - splits)) for j in range(len(gene_probs))])

        binom_scores = np.zeros((len(splits), k + 1))

        for i, p in enumerate(splits):
            if verbose:
                print('\r computing binom scores for bin {}/{}'.format(i, len(splits)), end=' ')

            # #print('using fast scores')
            # if fast_scores:
            signs = [1 if p * k <= j else -1 for j in np.arange(k + 1)]
            pmfs = binom.pmf(np.arange(k + 1), k, p)
            orderer = np.argsort(pmfs)

            pvals = np.cumsum(pmfs[orderer])


            if multi_correct:
                # use FWER to correct for testing many genes
                pvals_corrected = 1 - np.power(1 - pvals, n_tests)
                pvals_corrected[pvals < 1e-10] = taylor_exp(pvals[pvals < 1e-10], n_tests)  # more accurate

                pvals = pvals_corrected

            pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen
            binom_scores[i, orderer] = np.array(signs)[orderer] * -1 * np.log(pvals)

            # else:
            #     for j in range(k+1):
            #         alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
            #
            #         sign = 1 if p*k <= j else -1  # negative if lower than expected
            #
            #         binom_scores[i,j] = sign * -1*np.log(binom_test(j, n=k, p=p, alternative=alt))


    else:  # compute for all gene probs
        binom_scores = np.zeros((len(gene_probs), k + 1))
        gene_bins = np.array(range(len(gene_probs)))

        for i in range(len(gene_probs)):
            p = gene_probs[i]

            if verbose:
                print('\r computing binom scores for genes {}/{}'.format(i, len(gene_probs)), end=' ')
            if gene_probs[i] == 0 or gene_probs[i] == 1:  # no expression, no score
                continue

            signs = [1 if p * k <= j else -1 for j in np.arange(k + 1)]
            pmfs = binom.pmf(np.arange(k + 1), k, p)
            orderer = np.argsort(pmfs)

            pvals = np.cumsum(pmfs[orderer])

            if multi_correct:
                # use FWER to correct for testing many genes
                pvals_corrected = 1 - np.power(1 - pvals, n_tests)
                pvals_corrected[pvals < 1e-10] = taylor_exp(pvals[pvals < 1e-10], n_tests)  # more accurate

                pvals = pvals_corrected
            # pvals = 1-np.power(1-pvals, n_tests)

            pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen

            binom_scores[i, orderer] = np.array(signs)[orderer] * -1 * np.log(pvals)
            #binom_scores[i, orderer] = pvals #DEBUG ONLY TODO
            #
            # for j in range(k + 1):
            #
            #
            #
            #     p = gene_probs[i]
            #     alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
            #
            #     sign = 1 if p*k <= j else -1  # negative if lower than expected
            #
            #     binom_scores[i,j] = sign * -1*np.log(binom_test(j, n=k, p=p, alternative=alt))

    return (gene_bins, binom_scores)

    #return (gene_bins, binom_scores)


def info_score(X, nbhds, max_bins=float('inf'),
               entropy_normalize=False, fast_version=True, binom_scores=None, gene_bins=None,
               return_bin_info=False, verbose=True, n_tests='auto', model='wilcoxon',
               chunk_size=1000, **kwargs):
    """
    :param X: sparse count matrix
    :param nbhds: list with indices of nearest neighbors for each obs in X, e.g. from kneighbors() in sklearn
    :param max_bins: Resolution at which global gene probabilities are computed.
    if inf, genes get their own probabilities. Otherwise, the unit interval is split into max_bins pieces
    and they are rounded. This makes it faster with little performance difference
    :param return_all: if True, will also return global and local gene probabilities

    :param binom_scores: pass in binomial scores for each gene/bin, if pre-computed. Allows saving for future iterations.
    :param gene_bins: pass in gene bins from previous run. Speeds up iteration

    :param return_bin_info: for iteration: keep information about gene bins and binomial probs.
    :param fast_version: if True, use matrix multiplication instead of iteration. Fast, but memory-intensive.
    :return: dense matrix of gene/cell weightings.
    """



    if type(nbhds) is np.ndarray:
        nbhds = list(nbhds)

    k = len(nbhds[0])  # how many neighbors?

    if n_tests == 'auto':
        # determine by boostrapping
        n_tests = bootstrapped_ntests(X, k=k, model=model)

    wts = np.zeros((len(nbhds), X.shape[1]))  # too big for large data
    # nbhd_counts = np.zeros(X.shape)  # ditto
    # nbhd_sizes = [len(x) for x in NNs]

    # first compute frequencies of all genes:
    gene_probs = np.array((X > 0).sum(axis=0) / float(X.shape[0])).flatten()


    # frequencies of genes within neighborhoods
    nbhd_probs = np.zeros(X.shape)

    if model == 'ttest':
        data = np.ones(np.sum([len(x) for x in nbhds]))
        col_ind = [item for sublist in nbhds for item in sublist]
        row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

        # sparse adjacency matrix of NN graph
        nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

        # get mean gene expressions within each neighborhood; this matrix may be less sparse
        mean_nbhd_exprs = (nn_matrix * X).astype('int').multiply(1/nn_matrix.sum(axis=1)).tocsr()

        vars = np.zeros((len(nbhds), X.shape[1]))
        for i in range(len(nbhds)): # gotta go cell by cell
            nbrs = np.array(nbhds[i]).flatten()
            gene_diffs = np.power((X[nbrs,:].todense()-mean_nbhd_exprs[i,:].todense()),2) # diffs of gene expression
            vars[i,:] = gene_diffs.mean(axis=0)
        vars = csr_matrix(vars)

        global_means = np.tile(X.mean(axis=0), (len(nbhds),1))

        #sign is pos if mean is higher, negative otherwise.
        signs = 2*(mean_nbhd_exprs.todense() >= global_means).astype('int') - 1

        global_var = np.tile(np.var(X.todense(), axis=0), (len(nbhds),1))
        nobs_global = np.tile(X.shape[0], (len(nbhds), X.shape[1]))
        nobs_local = np.tile(k, (len(nbhds), X.shape[1]))

        wts = ttest_ind_from_stats(mean1=mean_nbhd_exprs.todense().flatten(),
                                              std1=np.array(np.sqrt(vars.todense()).flatten()),
                                              nobs1=np.array(nobs_local).flatten(),
                                              mean2=np.array(global_means).flatten(),
                                              std2=np.array(np.sqrt(global_var)).flatten(),
                                              nobs2=np.array(nobs_global).flatten()).pvalue.reshape((len(nbhds), X.shape[1]))

        np.nan_to_num(wts, copy=False, nan=1.0) # nans become pval 1

        wts[wts==0] = sys.float_info.min # remove zeros


        if n_tests>1:
            # use FWER to correct for testing many genes
            wts_corrected = 1 - np.power(1 - wts, n_tests)
            wts_corrected[wts < 1e-10] = taylor_exp(wts[wts < 1e-10], n_tests)  # more accurate

            wts = wts_corrected
        else:
            wts_corrected = wts

        wts = -1*np.log(wts) # convert to info

        np.nan_to_num(wts, copy=False, nan=1.0)  # nans become pval 1

        wts = np.multiply(signs, wts) # negative if underexpressed





        return(csr_matrix(wts))

    #TODO TODO add signs

    elif model == 'wilcoxon':
        from scipy.stats import rankdata

        def fastRank(array):
            temp = array.argsort(axis=0)
            ranks = np.zeros(temp.shape)

            rows = temp.transpose().flatten()
            cols = np.repeat(np.arange(temp.shape[1]), temp.shape[0])
            ranks[rows, cols] = np.array(list(np.arange(temp.shape[0])) * temp.shape[1])
            return (ranks)

        # Wilcoxon rank sum testa
        #overall_exprs = X.todense().transpose().tolist()

        n_genes = X.shape[1]
        chunk_ends = [0] + list(np.arange(chunk_size, n_genes, chunk_size))
        chunk_ends.append(n_genes)
        gene_idxs = np.array(list(range(n_genes)))
        gene_chunks = [np.array(gene_idxs[chunk_ends[i]:chunk_ends[i + 1]]) for i in range(len(chunk_ends) - 1)]

        wt_blocks = [] # list of sparse matrices to concatenate horizontally

        # make nbhd adjacency matrix
        data = np.ones(np.sum([len(x) for x in nbhds]))
        col_ind = [item for sublist in nbhds for item in sublist]
        row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

        # sparse adjacency matrix of NN graph
        nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

        for i,chunk in enumerate(gene_chunks):
            if verbose:
                print('chunk {}/{}'.format(i+1,len(gene_chunks)), end='\r')

            X_chunk = X[:,chunk]

            wts = rankdata(X_chunk.todense(), axis=0) # gene rankings

            wts = nn_matrix @ wts  # nbhd_ranksums; only want to store one big matrix

            n1 = k
            n2 = X_chunk.shape[0] - k
            sd = np.sqrt(n1 * n2 * (n1 + n2 + 1) / 12.0)
            meanrank = n1 * n2 / 2.0

            # #sign is pos if mean rank is higher than average, negative otherwise.
            # signs = 2*(wts >= meanrank).astype('int') - 1

            wts = wts - ((n1 * (n1 + 1)) / 2.0)  # calc U for x, u1

            is_neg = (wts<meanrank) # remember where it was negative

            wts = np.maximum(wts, n1 * n2 - wts)  # bigu

            wts = ((wts - meanrank) / sd) # z values

            wts = 2 * norm.sf(np.abs(wts)) #p values
            #
            # for i in range(len(nbhds)):
            #     print('cell {}/{}'.format(i+1,len(nbhds)+1), end='\r')
            #     #gene_exprs = X[nbhds[i],:].todense()
            #     #all_exprs = np.vstack((gene_exprs, X.todense()))
            #
            #     nbhd_ranks = gene_rankings[nbhds[i],:]
            #
            #     ranksums = np.sum(nbhd_ranks, axis=0)
            #     n1 = k
            #     n2 = X.shape[0]-k
            #     #
            #     # ranks = fastRank(all_exprs.A)
            #     #
            #     # n1 = k
            #     # n2 = X.shape[0]
            #     # ranksums = np.sum(ranks[:k, :], axis=0)
            #     u1 = ranksums - ((n1 * (n1 + 1)) / 2.0)  # calc U for x
            #     u2 = n1 * n2 - u1  # remainder is U for y
            #
            #     sd = np.sqrt(n1 * n2 * (n1 + n2 + 1) / 12.0)
            #     meanrank = n1 * n2 / 2.0
            #
            #     bigu = np.maximum(u1, u2)
            #    # wts[i,:] = bigu
            #
            #     z = ((bigu - meanrank) / sd).flatten()
            #     p = 2 * norm.sf(abs(z))
            #     wts[i,:] = p
            #
            #

                # gene_exprs = X[nbhds[i],:].todense().transpose().tolist() # list of gene expression vectors for each nbr
                # for j,local_expr in enumerate(gene_exprs):
                #     global_expr = overall_exprs[j]
                #     wts[i,j] = mannwhitneyu(x=local_expr, y=global_expr, alternative='two-sided', use_continuity=False)[1]
            if n_tests>1:
                # use FWER to correct for testing many genes
                wts[wts> 1e-10] = 1-np.power(1-wts[wts>1e-10], n_tests)
                wts[wts<=1e-10] = taylor_exp(wts[wts <= 1e-10], n_tests)
                # wts_corrected = 1 - np.power(1 - wts, n_tests)
                # wts_corrected[wts < 1e-10] = taylor_exp(wts[wts < 1e-10], n_tests)  # more accurate
                #
                # wts = wts_corrected

            wts = -1*np.log(wts) # convert to info scores

            #sign them
            wts[is_neg] *= -1
            #wts = np.multiply(signs, wts)

            wt_blocks.append(csr_matrix(wts))

        return(hstack(wt_blocks))

    elif model == 'log_likelihood':
        means = X.mean(axis=0)
        variances = sparse_vars(X, axis=0)/float(k)

        data = np.ones(np.sum([len(x) for x in nbhds]))
        col_ind = [item for sublist in nbhds for item in sublist]
        row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

        # sparse adjacency matrix of NN graph
        nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

        nbhd_means = ((nn_matrix * X)/float(k)).todense()

        print(np.min(variances))
        wts = 1 / 2. * np.log(2 * np.pi) + np.power((nbhd_means - means), 2) / (2 * variances) + np.log(variances)/2.

        signs = 2*(nbhd_means >= means).astype('int') - 1

        wts = np.multiply(wts, signs)

        return(csr_matrix(wts))



    elif model == 'binomial':
        X = csr_matrix((X > 0).astype('float'))  # convert to sparse binarized matrix
        if binom_scores is None or gene_bins is None:
            if n_tests is None:
                n_tests = X.shape[1]  # multi-correct per cell
            gene_bins, binom_scores = get_binom_scores(gene_probs, k, max_bins=max_bins,
                                                       verbose=verbose, n_tests=n_tests, **kwargs)

        if fast_version:
            # compute significance of gene expression in each cell's neighborhood
            # first convert neighborhood to sparse matrix
            data = np.ones(np.sum([len(x) for x in nbhds]))
            col_ind = [item for sublist in nbhds for item in sublist]
            row_ind = [i for i, sublist in enumerate(nbhds) for item in sublist]

            # sparse adjacency matrix of NN graph
            nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(len(nbhds), X.shape[0]))

            # get gene expressions within each neighborhood; this matrix may be less sparse
            nbhd_exprs = (nn_matrix * X).astype('int').todense()

            # # extract locations and values of nonzero nbhd expressions.
            # rows, cols = nbhd_exprs.nonzero()
            # exprs = nbhd_exprs.data

            # apply binomial scores
            rows, cols = np.indices((len(nbhds), X.shape[1]))
            rows = rows.flatten()
            cols = cols.flatten()

            wts = binom_scores[gene_bins[cols], np.array(nbhd_exprs[rows, cols]).flatten()].reshape((len(nbhds), X.shape[1]))

        else:
            for i in range(len(nbhds)):
                if verbose:
                    if i < len(nbhds) - 1:
                        print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         ')
                    else:
                        print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         \n')

                nnbhd = X[nbhds[i], :]

                nbhd_size = len(nbhds[i])
                nbhd_gene_counts = np.array((nnbhd > 0).sum(axis=0)).flatten()

                nbhd_probs[i, :] = nbhd_gene_counts / nbhd_size

                if max_bins < float('inf'):
                    # look up the binomial score in the nearest bins

                    # gene_scores = [binom_scores[gene_bins[j], count] for j, count in enumerate(nbhd_gene_counts)]
                    gene_scores = binom_scores[gene_bins, nbhd_gene_counts]

                else:
                    gene_scores = [binom_scores[j, count] for j, count in enumerate(nbhd_gene_counts)]

                wts[i, :] = gene_scores
                # expected_vals = nbhd_size * gene_probs
                # wts[i, :] = -1*np.log(gene_scores) * (2*(nbhd_gene_counts > expected_vals)-1)

    if entropy_normalize:  # divide each column by the entropy of the corresponding gene
        gene_entropies = -1 * (np.multiply(gene_probs, np.log(gene_probs))
                               + np.multiply((1 - gene_probs), np.log(1 - gene_probs)))
        gene_entropies[np.logical_not(np.isfinite(gene_entropies))] = float(
            'inf')  # zeros out non-expressed or everywhere-expressed genes
        wts = np.divide(wts, gene_entropies)

    # if return_all:
    #     return (wts, gene_probs, nbhd_probs)
    wts = csr_matrix(wts)
    if model == 'binomial' and return_bin_info:  # for iteration
        return (wts, gene_bins, binom_scores)
    else:
        return (wts)

    #
    # # pre-compute binomial scores to avoid doing it for each cell/gene
    # if max_bins < float('inf') and binom_scores is None:
    #     # split interval into max_bins bins
    #     splits = np.arange(0, 1, 1/max_bins)[1:] # don't allow a gene to have 0 prob...
    #     gene_bins = np.array([np.argmin(np.abs(gene_probs[j] - splits)) for j in range(X.shape[1])])
    #
    #

    #     binom_scores = np.zeros((len(splits), k+1))
    #     for i, p in enumerate(splits):
    #         print('\r computing binom scores for bin {}/{}'.format(i, len(splits)), end=' ')
    #         for j in range(k+1):
    #             alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
    #
    #             if p_val: # information of any occurrence more extreme
    #                 binom_scores[i,j] = binom_test(j, n=k, p=p, alternative=alt)
    #             else: #information of this actual occurrence
    #                 val = -1*(j*np.log(p)+(k-j)*np.log(1-p))
    #                 if np.isfinite(val): # zero out nans
    #                     binom_scores[i,j] = -1*(j*np.log(p)+(k-j)*np.log(1-p))

    #
    # elif binom_scores is None: # compute for all gene probs
    #     binom_scores = np.zeros((X.shape[1], k + 1))
    #     for i in range(X.shape[1]):
    #         print('\r computing binom scores for genes {}/{}'.format(i, X.shape[1]), end=' ')
    #         for j in range(k + 1):
    #             p=gene_probs[i]
    #             alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
    #             if p_val: # information of any occurrence more extreme
    #                 binom_scores[i,j] = binom_test(j, n=k, p=p, alternative=alt)
    #             else: #information of this actual occurrence
    #                 val = -1*(j*np.log(p)+(k-j)*np.log(1-p))
    #                 if np.isfinite(val): # zero out nans
    #                     binom_scores[i,j] = -1*(j*np.log(p)+(k-j)*np.log(1-p))
